from typing import List

from pydantic import BaseModel, Field

from galileo_core.schemas.shared.scorers.base_configs import GeneratedScorerConfig, RegisteredScorerConfig
from galileo_core.schemas.shared.scorers.scorers import GalileoScorer


class ScorerConfiguration(BaseModel):
    scorers: List[GalileoScorer] = Field(default_factory=list, description="List of Galileo scorers to enable.")
    registered_scorers: List[RegisteredScorerConfig] = Field(
        default_factory=list, description="List of registered scorers to enable."
    )
    generated_scorers: List[GeneratedScorerConfig] = Field(
        default_factory=list, description="List of generated scorers to enable."
    )



from enum import Enum
from typing import Union

from pydantic import BaseModel, Field

from galileo_core.schemas.shared.metric import MetricValueType


class RuleOperator(str, Enum):
    gt = "gt"
    lt = "lt"
    gte = "gte"
    lte = "lte"
    eq = "eq"
    neq = "neq"
    contains = "contains"
    all = "all"
    any = "any"
    empty = "empty"
    not_empty = "not_empty"


class Rule(BaseModel):
    metric: str = Field(description="Name of the metric.")
    operator: RuleOperator = Field(description="Operator to use for comparison.")
    target_value: Union[str, float, int, list, None] = Field(
        description="Value to compare with for this metric (right hand side)."
    )

    def evaluate(self, value: MetricValueType) -> bool:
        if value is not None:
            if isinstance(value, (float, int)) and isinstance(self.target_value, (float, int)):
                if self.operator == RuleOperator.gt:
                    return value > self.target_value
                elif self.operator == RuleOperator.lt:
                    return value < self.target_value
                elif self.operator == RuleOperator.gte:
                    return value >= self.target_value
                elif self.operator == RuleOperator.lte:
                    return value <= self.target_value
                elif self.operator == RuleOperator.eq:
                    return value == self.target_value
                elif self.operator == RuleOperator.neq:
                    return value != self.target_value
            elif isinstance(value, str):
                if isinstance(self.target_value, (str)):
                    if self.operator == RuleOperator.eq:
                        return value == self.target_value
                    elif self.operator == RuleOperator.neq:
                        return value != self.target_value
                elif isinstance(self.target_value, list):
                    if self.operator == RuleOperator.any:
                        return any([t == value for t in self.target_value])
            elif isinstance(value, list):
                if isinstance(self.target_value, str):
                    if self.operator == RuleOperator.contains:
                        return self.target_value in value
                elif isinstance(self.target_value, list):
                    if self.operator == RuleOperator.all:
                        return all([v in value for v in self.target_value])
                    elif self.operator == RuleOperator.any:
                        return any([v in value for v in self.target_value])
                elif self.target_value is None:
                    if self.operator == RuleOperator.empty:
                        return len(value) == 0
                    elif self.operator == RuleOperator.not_empty:
                        return len(value) > 0

        return False



from json import dumps
from typing import Any, Dict, List, Literal, Optional, Sequence, Union, cast

from pydantic import ConfigDict, Field, field_validator
from pydantic_core import ValidationError
from typing_extensions import Annotated

from galileo_core.schemas.logging.llm import Message, MessageRole
from galileo_core.schemas.logging.step import BaseStep, Metrics, StepType
from galileo_core.schemas.shared.document import Document
from galileo_core.utils.json import PydanticJsonEncoder

LlmSpanAllowedInputType = Union[Sequence[Message], Message, str, Dict[str, Any], Sequence[Dict[str, Any]]]
LlmSpanAllowedOutputType = Union[Message, str, Dict[str, Any]]


class StepWithChildSpans(BaseStep):
    spans: List["Span"] = Field(default_factory=list, description="Child spans.")

    def add_child_spans(self, spans: Sequence["Span"]) -> None:
        self.spans.extend(spans)

    def add_child_span(self, span: "Span") -> None:
        self.add_child_spans([span])


class BaseWorkflowSpan(BaseStep):
    type: Literal[StepType.workflow] = Field(
        default=StepType.workflow, description=BaseStep.model_fields["type"].description
    )


class WorkflowSpan(BaseWorkflowSpan, StepWithChildSpans):
    pass


class LlmMetrics(Metrics):
    num_input_tokens: Optional[int] = Field(default=None, description="Number of input tokens.")
    num_output_tokens: Optional[int] = Field(default=None, description="Number of output tokens.")
    num_total_tokens: Optional[int] = Field(default=None, description="Total number of tokens.")
    time_to_first_token_ns: Optional[int] = Field(
        default=None,
        description="Time until the first token was generated in nanoseconds.",
    )

    model_config = ConfigDict(extra="allow")


class LlmSpan(BaseStep):
    type: Literal[StepType.llm] = Field(default=StepType.llm, description=BaseStep.model_fields["type"].description)
    input: Sequence[Message] = Field(description=BaseStep.model_fields["input"].description)
    output: Message = Field(description=BaseStep.model_fields["output"].description)
    metrics: LlmMetrics = Field(default_factory=LlmMetrics, description=BaseStep.model_fields["metrics"].description)
    tools: Optional[Sequence[Dict[str, Any]]] = Field(
        default=None,
        description="List of available tools passed to the LLM on invocation.",
    )
    model: Optional[str] = Field(default=None, description="Model used for this span.")
    temperature: Optional[float] = Field(default=None, description="Temperature used for generation.")
    finish_reason: Optional[str] = Field(default=None, description="Reason for finishing.")

    @classmethod
    def _convert_dict_to_message(cls, value: Dict[str, Any], default_role: MessageRole = MessageRole.user) -> Message:
        """
        Converts a dict into a Message object.
        Will dump the dict to a json string if it unable to be deserialized into a Message object.

        Args:
            value (Dict[str, Any]): The dict to convert.
            default_role (Optional[MessageRole], optional): The role to use if the dict does not contain a role. Defaults to MessageRole.user.

        Returns:
            Message: The converted Message object.
        """
        try:
            return Message.model_validate(value)
        except ValidationError:
            return Message(content=dumps(value), role=default_role)

    @field_validator("tools", mode="after")
    def validate_tools_serializable(cls, val: Optional[Sequence[Dict[str, Any]]]) -> Optional[Sequence[Dict[str, Any]]]:
        # Make sure we can dump input/output to json string.
        dumps(val, cls=PydanticJsonEncoder)
        return val

    @field_validator("input", mode="before")
    def convert_input(cls, value: LlmSpanAllowedInputType) -> Sequence[Message]:
        """Converts various input types into a standardized list of Message objects."""
        if isinstance(value, Sequence) and all(isinstance(item, Message) for item in value):
            return cast(Sequence[Message], value)
        if isinstance(value, Sequence) and all(isinstance(item, Dict) for item in value):
            return [
                cls._convert_dict_to_message(value=cast(Dict[str, Any], item), default_role=MessageRole.user)
                for item in value
            ]
        if isinstance(value, str):
            return [Message(role=MessageRole.user, content=value)]
        if isinstance(value, Message):
            return [value]
        if isinstance(value, Dict):
            return [cls._convert_dict_to_message(value=value, default_role=MessageRole.user)]
        raise ValueError("LLM span input must be a Message, a list of Messages, a dict, a list of dicts, or a string.")

    @field_validator("output", mode="before")
    def convert_output(cls, value: LlmSpanAllowedOutputType) -> Message:
        """Converts various output types into a standardized Message object."""
        if isinstance(value, Message):
            return value
        if isinstance(value, str):
            return Message(role=MessageRole.assistant, content=value)
        if isinstance(value, Dict):
            return cls._convert_dict_to_message(value=value, default_role=MessageRole.assistant)
        raise ValueError("LLM span output must be a Message, a string, or a dict.")


class RetrieverSpan(BaseStep):
    type: Literal[StepType.retriever] = Field(
        default=StepType.retriever, description=BaseStep.model_fields["type"].description
    )
    input: str = Field(description=BaseStep.model_fields["input"].description)
    output: List[Document] = Field(description=BaseStep.model_fields["output"].description)

    @field_validator("output", mode="before")
    def set_output(cls, value: Union[List[Dict[str, str]], List[Document]]) -> List[Document]:
        if isinstance(value, list):
            if all(isinstance(doc, dict) for doc in value):
                parsed = [Document.model_validate(doc) for doc in value]
            elif all(isinstance(doc, Document) for doc in value):
                parsed = [Document.model_validate(doc) for doc in value]
            else:
                raise ValueError("Retriever output must be a list of dicts, or a list of Documents.")
            return parsed
        raise ValueError("Retriever output must be a list of dicts or a list of Documents.")


class ToolSpan(BaseStep):
    type: Literal[StepType.tool] = Field(default=StepType.tool, description=BaseStep.model_fields["type"].description)
    input: str = Field(description=BaseStep.model_fields["input"].description)
    output: Optional[str] = Field(default=None, description=BaseStep.model_fields["output"].description)
    tool_call_id: Optional[str] = Field(default=None, description="ID of the tool call.")


Span = Annotated[Union[WorkflowSpan, LlmSpan, RetrieverSpan, ToolSpan], Field(discriminator="type")]

StepWithChildSpans.model_rebuild()

SpanStepTypes = [step_type.value for step_type in StepType if step_type != StepType.trace]



from abc import ABC
from logging import WARNING
from typing import Any, Dict, Optional

from galileo_core.helpers.logger import logger


class BaseGalileoException(Exception, ABC):
    """Base exception for all exceptions in galileo."""

    LOG_LEVEL: int = WARNING

    def __init__(self, message: str, logging_extra: Optional[Dict[str, Any]] = None) -> None:
        logger.log(self.LOG_LEVEL, message, extra=logging_extra)
        self.message: str = message
        super().__init__(self.message)



from enum import Enum
from typing import Any, Dict, Generator, List, Optional

from pydantic import BaseModel, RootModel, model_validator


class MessageRole(str, Enum):
    agent = "agent"
    assistant = "assistant"
    developer = "developer"
    function = "function"
    system = "system"
    tool = "tool"
    user = "user"


class Message(BaseModel):
    content: str
    role: MessageRole
    tool_call_id: Optional[str] = None
    tool_calls: Optional[List["ToolCall"]] = None

    @model_validator(mode="before")
    def _allow_null_content_with_tool_calling(cls, data: Dict[str, Any]) -> Dict[str, Any]:
        # Some APIs (like OpenAI) often set content to None when there are tool calls.
        # This is a workaround to allow for that case without changing the type of the field.
        # TODO: Consider making the content field nullable.
        if data.get("content") is None:
            if data.get("tool_calls") is None:
                raise ValueError("at most one of 'content' and 'tool_calls' can be None, but both were None")
            # Deep copy and preserve key order
            data = {k: "" if k == "content" else v for k, v in data.items()}
        return data


class Messages(RootModel[List[Message]]):
    def __len__(self) -> int:
        return len(self.root)

    def __iter__(self) -> Generator[Message, None, None]:  # type: ignore[override]
        yield from self.root

    def __getitem__(self, item: int) -> Message:
        return self.root[item]


class ToolCall(BaseModel):
    id: str
    function: "ToolCallFunction"


class ToolCallFunction(BaseModel):
    name: str
    arguments: str



from datetime import datetime, timezone
from enum import Enum
from json import dumps
from typing import Any, Dict, List, Optional, Sequence, Union

from pydantic import BaseModel, ConfigDict, Field, ValidationError, field_validator
from pydantic_core.core_schema import ValidationInfo

from galileo_core.schemas.logging.llm import Message
from galileo_core.schemas.shared.document import Document
from galileo_core.utils.json import PydanticJsonEncoder

StepAllowedInputType = Union[str, Sequence[Message]]
StepAllowedOutputType = Union[None, str, Message, Sequence[Document]]


class StepType(str, Enum):
    llm = "llm"
    retriever = "retriever"
    tool = "tool"
    workflow = "workflow"
    trace = "trace"


class Metrics(BaseModel):
    duration_ns: Optional[int] = Field(
        default=None, description="Duration of the trace or span in nanoseconds.  Displayed as 'Latency' in Galileo."
    )

    model_config = ConfigDict(extra="allow")


class BaseStep(BaseModel):
    type: StepType = Field(description="Type of the trace or span.")
    input: StepAllowedInputType = Field(description="Input to the trace or span.", union_mode="left_to_right")
    output: StepAllowedOutputType = Field(
        default=None, description="Output of the trace or span.", union_mode="left_to_right"
    )
    name: str = Field(default="", description="Name of the trace or span.", validate_default=True)
    created_at: datetime = Field(
        title="Created",
        default_factory=lambda: datetime.now(tz=timezone.utc),
        description="Timestamp of the trace or span's creation.",
    )
    user_metadata: Dict[str, str] = Field(
        default_factory=dict, description="Metadata associated with this trace or span."
    )
    tags: List[str] = Field(default_factory=list, description="Tags associated with this trace or span.")
    status_code: Optional[int] = Field(
        default=None, description="Status code of the trace or span. Used for logging failure or error states."
    )
    metrics: Metrics = Field(default_factory=Metrics, description="Metrics associated with this trace or span.")

    model_config = ConfigDict(validate_assignment=True)

    def __init__(self, **data: Any):
        for k, v in list(data.items()):
            if v is None:
                del data[k]
        super().__init__(**data)

    @field_validator("name", mode="before")
    def set_name(cls, value: Optional[str], info: ValidationInfo) -> str:
        if value is not None:
            return value
        if "type" in info.data:
            return info.data["type"]
        raise ValidationError("could not set step name from type since type is missing")

    @field_validator("input", mode="after")
    def validate_input_serializable(cls, val: StepAllowedInputType) -> StepAllowedInputType:
        # Make sure we can dump input/output to json string.
        dumps(val, cls=PydanticJsonEncoder)
        return val

    @field_validator("output", mode="after")
    def validate_output_serializable(cls, val: StepAllowedOutputType) -> StepAllowedOutputType:
        # Make sure we can dump input/output to json string.
        dumps(val, cls=PydanticJsonEncoder)
        return val



from enum import Enum

from pydantic import UUID4, BaseModel, Field, field_validator

from galileo_core.utils.name import ts_name

DEFAULT_PROJECT_NAME = "project"


class ProjectType(str, Enum):
    training_inference = "training_inference"
    prompt_evaluation = "prompt_evaluation"
    llm_monitor = "llm_monitor"
    protect = "protect"
    gen_ai = "gen_ai"


class CreateProjectRequest(BaseModel):
    name: str = Field(default=DEFAULT_PROJECT_NAME, validate_default=True)
    type: ProjectType

    @field_validator("name", mode="before")
    def generate_name(cls, value: str) -> str:
        if value == DEFAULT_PROJECT_NAME:
            value = ts_name(prefix=value)
        return value


class ProjectResponse(CreateProjectRequest):
    id: UUID4



from typing import Any, Dict, Union

from pydantic import BaseModel, ConfigDict, Field, field_validator

ChunkMetaDataValueType = Union[bool, str, int, float]


class Document(BaseModel):
    content: str = Field(description="Content of the document.", validation_alias="page_content")
    metadata: Dict[str, ChunkMetaDataValueType] = Field(default_factory=dict, validate_default=True)

    model_config = ConfigDict(populate_by_name=True, extra="forbid")

    @field_validator("metadata", mode="before")
    def filter_metadata(cls, metadata: Any) -> Dict[str, ChunkMetaDataValueType]:
        # We don't want to throw an error due to unexpected metadata, so we filter it out.
        if not isinstance(metadata, Dict):
            return dict()
        return {key: val for key, val in metadata.items() if isinstance(val, (bool, str, int, float))}



from typing import Generator, List

from pydantic import RootModel, field_validator

from galileo_core.schemas.shared.message import Message
from galileo_core.schemas.shared.message_role import MessageRole


class Messages(RootModel[List[Message]]):
    def __len__(self) -> int:
        return len(self.root)

    def __iter__(self) -> Generator[Message, None, None]:  # type: ignore[override]
        yield from self.root

    def __getitem__(self, item: int) -> Message:
        return self.root[item]

    @field_validator("root", mode="after")
    def system_message_first(cls, messages: List[Message]) -> List[Message]:
        for message in messages[1:]:
            if message.role == MessageRole.system:
                raise ValueError("There can only be 1 system message and it must be the first message.")
        return messages



from enum import Enum


class ConfigEnvironmentVariables(str, Enum):
    console_url = "GALILEO_CONSOLE_URL"
    username = "GALILEO_USERNAME"
    password = "GALILEO_PASSWORD"
    sso_id_token = "GALILEO_SSO_ID_TOKEN"
    sso_provider = "GALILEO_SSO_PROVIDER"
    api_key = "GALILEO_API_KEY"
    jwt_token = "GALILEO_JWT_TOKEN"



import inspect
from ast import parse
from importlib.util import module_from_spec, spec_from_file_location
from itertools import chain
from traceback import format_exc
from types import ModuleType
from typing import Any, Callable, Dict, List, Optional, Union, get_args
from uuid import UUID

from galileo_core.exceptions.execution import ExecutionError
from galileo_core.helpers.logger import logger
from galileo_core.schemas.shared.metric import UserMetricType
from galileo_core.schemas.shared.scorers.base_configs import ScoreType
from galileo_core.schemas.shared.scorers.chain_aggregation import ChainAggregationStrategy
from galileo_core.schemas.shared.workflows.node_type import NodeType

DEFAULT_SCOREABLE_NODE_TYPES: List[NodeType] = [NodeType.llm, NodeType.chat]
SCORER_FN_REQUIRED_ARGS = ["kwargs"]
AGGREGATOR_FN_REQUIRED_ARGS = ["scores"]


def default_scorer_fn(*, index: Union[int, UUID], **kwargs: Any) -> UserMetricType:
    """
    Default scorer function that is used when the user-registered scorer does not provide a scorer function.

    All args are keyword-only.

    Parameters
    ----------
    index : Union[int, UUID]
        Index of the node.

    Returns
    -------
    UserMetricType
        Score.
    """
    return None


def default_aggregator_fn(*, scores: List[UserMetricType], **kwargs: Any) -> Dict[str, UserMetricType]:
    """
    Default aggregator function that is used when the user-registered scorer does not provide an aggregator function.

    All args are keyword-only.

    Parameters
    ----------
    scores : List[UserMetricType]
        List of scores from the scorer function.

    Returns
    -------
    Dict[str, Any]
        Aggregated scores.
    """
    return dict()


def validate_registered_scorer(file_path: str, extra: dict, exec_module: bool = True) -> ModuleType:
    """
    Validate the registered scorer file.

    Args:
        file_path: The path to the registered scorer Python file.
        extra: Extra information to log.

    Returns:
        The module object of the registered scorer file.

    Raises:
        ExecutionError:
            If the scorer file contains invalid Python code or if the scorer function is not found.
    """
    with open(file_path) as open_file:
        try:
            content = open_file.read()

            parse(content)
        except SyntaxError as exception:
            raise ExecutionError(
                f"Invalid Python code in scorer file. SyntaxError when parsing: {exception.msg}.", extra
            )

    spec = spec_from_file_location("scorer", file_path)
    if spec is None or spec.loader is None:
        raise ExecutionError("Error loading registered scorer module spec.", extra)
    user_module = module_from_spec(spec)

    if exec_module:  # Allow this to be turned off (for G2.0 sandbox mode)
        try:
            spec.loader.exec_module(user_module)
        except Exception:
            raise ExecutionError(f"Error while loading your custom metric code: \n\n {format_exc(chain=False)}.", extra)

    if scorer_fn := getattr(user_module, "scorer_fn", None):
        if not callable(scorer_fn):
            raise ExecutionError("scorer_fn must be a function.", extra)

    logger.info("Successfully validated scorer file.", extra=extra)
    return user_module


def validate_score_type_fn(user_module: ModuleType, extra: dict) -> type:
    """
    Validate the score_type function in the registered scorer file.

    Args:
        user_module: The module object of the registered scorer file.
        extra: Extra information to log.

    Returns:
        The score type returned by the score_type function.

    Raises:
        ExecutionError: If the score_type function is not a function or does not return a valid score type.
    """
    if score_type_fn := getattr(user_module, "score_type", None):
        if not callable(score_type_fn):
            raise ExecutionError("score_type must be a function.", extra)

        score_type = score_type_fn()
        if score_type not in get_args(ScoreType):
            raise ExecutionError(f"score_type must return one of {get_args(ScoreType)}.", extra)

        return score_type

    return float


def validate_scoreable_node_types_fn(user_module: ModuleType, extra: dict) -> List[NodeType]:
    """
    Validate the scoreable_node_types function in the registered scorer file.

    Args:
        user_module: The module object of the registered scorer file.
        extra: Extra information to log.

    Returns:
        The list of scorable node types returned by the scoreable_node_types function.

    Raises:
        ExecutionError:
            If the scoreable_node_types function is not a function
            or does not return a valid list of scorable node types.
    """
    if scoreable_node_types_fn := getattr(user_module, "scoreable_node_types_fn", None):
        if not callable(scoreable_node_types_fn):
            raise ExecutionError("scoreable_node_types_fn must be a function.", extra)

        scoreable_node_types = scoreable_node_types_fn()
        if not isinstance(scoreable_node_types, list):
            raise ExecutionError("scoreable_node_types_fn must return a list.", extra)

        for value in scoreable_node_types:
            if value not in NodeType.__members__:
                raise ExecutionError(
                    f"Node type {value} is invalid. Must be one of {', '.join(NodeType.__members__.keys())}.", extra
                )

        return scoreable_node_types

    return DEFAULT_SCOREABLE_NODE_TYPES


def validate_include_llm_credentials(user_module: ModuleType, extra: dict) -> bool:
    """
    Validate the include_llm_credentials function in the registered scorer file.

    Args:
        user_module: The module object of the registered scorer file.
        extra: Extra information to log.

    Returns:
        The value returned by the include_llm_credentials function.

    Raises:
        ExecutionError: If the include_llm_credentials is not a boolean.
    """
    if include_llm_credentials := getattr(user_module, "include_llm_credentials", False):
        if not isinstance(include_llm_credentials, bool):
            raise ExecutionError("include_llm_credentials must be a boolean.", extra)
    return include_llm_credentials or False


def validate_chain_aggregation(
    user_module: ModuleType, score_type: type, extra: dict
) -> Optional[ChainAggregationStrategy]:
    """
    Validate the chain_aggregation strategy in the registered scorer file.

    Args:
        user_module: The module object of the registered scorer file.
        score_type: The score type returned by the score_type function.
        Used to determine the default chain aggregation strategy if not provided.
        extra: Extra information to log.

    Returns:
        The value returned by the chain_aggregation strategy.

    Raises:
        ExecutionError: If the chain_aggregation is not a string or is an invalid string.
    """
    if chain_aggregation := getattr(user_module, "chain_aggregation", None):
        if not isinstance(chain_aggregation, str):
            raise ExecutionError("chain_aggregation must be a string.", extra)
        chain_aggregation = chain_aggregation.lower()
        if chain_aggregation not in chain.from_iterable(
            [strategy, strategy.name, strategy.value] for strategy in ChainAggregationStrategy
        ):
            raise ExecutionError("chain_aggregation is not a valid strategy.", extra)
        return ChainAggregationStrategy(chain_aggregation)
    if score_type in (int, float, bool):
        return ChainAggregationStrategy.average
    return None


def validate_scorer_fn(user_module: ModuleType, extra: dict) -> Optional[Callable]:
    """
    Validate the scorer_fn function in the registered scorer file.

    Args:
        user_module: The module object of the registered scorer file.
        extra: Extra information to log.

    Raises:
        ExecutionError: If the scorer_fn function is not a function or does
        not have the required arguments.
    """
    if scorer_fn := getattr(user_module, "scorer_fn", None):
        if not callable(scorer_fn):
            raise ExecutionError("scorer_fn must be a function.", extra)

        params = inspect.signature(scorer_fn).parameters

        for arg in SCORER_FN_REQUIRED_ARGS:
            if arg not in params:
                raise ExecutionError(f"scorer_fn must have the argument '{arg}'.", extra)

        return scorer_fn

    return None


def validate_aggregator_fn(user_module: ModuleType, extra: dict) -> Optional[Callable]:
    """
    Validate the aggregator_fn function in the registered scorer file.

    Args:
        user_module: The module object of the registered scorer file.
        extra: Extra information to log.

    Raises:
        ExecutionError: If the aggregator_fn function is not a function or
        does not have the required arguments.
    """
    if aggregator_fn := getattr(user_module, "aggregator_fn", None):
        if not callable(aggregator_fn):
            raise ExecutionError("aggregator_fn must be a function.", extra)

        params = inspect.signature(aggregator_fn).parameters

        for arg in AGGREGATOR_FN_REQUIRED_ARGS:
            if arg not in params:
                raise ExecutionError(f"aggregator_fn must have the argument '{arg}'.", extra)

        return aggregator_fn

    return None


def validate_node_type_input(
    scorer_fn: Callable, node_type: NodeType, expected_score_type: type, data: dict, extra: dict
) -> Optional[ScoreType]:
    """
    Test the scorer_fn function by running it with sample input data.

    Args:
        scorer_fn: The scorer function to test.
        node_type: The node type to test the scorer function with.
        expected_score_type: The expected score type.
        data: The input data to test the scorer function with.
        extra: Extra information to log.

    Returns:
        The score type returned by the scorer function.

    Raises:
        ExecutionError: If the scorer function raises an error.
    """

    try:
        score = scorer_fn(node_type=node_type, **data)
    except Exception as e:
        msg = f"Error running scorer function on node type '{node_type.name}': {e}"
        logger.error(msg, exc_info=True)
        raise ExecutionError(msg, extra)

    try:
        return coerce_type(score, expected_score_type)
    except ExecutionError as e:
        msg = f"Error running scorer function on node type '{node_type.name}': {e}"
        logger.error(msg, exc_info=True)
        raise ExecutionError(msg, extra)


def coerce_type(value: Any, expected_score_type: type) -> Optional[ScoreType]:
    """
    Coerce the value to the expected type.

    Parameters
    ----------
    value : Any
        Value to coerce.
    expected_score_type : Type
        Expected type.

    Returns
    -------
    UserMetricType
        Coerced value.

    Raises
    ------
    ExecutionError
        If the value cannot be coerced.
    """
    if value is None:
        return None

    try:
        return expected_score_type(value)
    except (ValueError, TypeError):
        raise ExecutionError(f"Failed to coerce value '{value}' to type {expected_score_type.__name__}.")



from typing import Dict, List, Optional

from pydantic import BaseModel, Field

from galileo_core.schemas.shared.traces.types import (
    LlmSpan,
    RetrieverSpan,
    StepWithChildSpans,
    ToolSpan,
    Trace,
    WorkflowSpan,
)
from galileo_core.schemas.shared.workflows.step import LlmStepAllowedIOType, RetrieverStepAllowedOutputType, StepIOType


class Traces(BaseModel):
    traces: List[Trace] = Field(default_factory=list, description="List of traces.")
    current_parent: Optional[StepWithChildSpans] = Field(default=None, description="Current parent context.")

    def add_trace(
        self,
        input: StepIOType,
        output: Optional[StepIOType] = None,
        name: Optional[str] = None,
        duration_ns: Optional[int] = None,
        created_at_ns: Optional[int] = None,
        metadata: Optional[Dict[str, str]] = None,
        ground_truth: Optional[str] = None,
    ) -> Trace:
        """
        Create a new trace and add it to the list of traces.
        Simple usage:
        ```
        my_traces.add_trace("input")
        my_traces.add_llm_span("input", "output", model="<my_model>")
        my_traces.conclude("output")
        ```
        Parameters:
        ----------
            input: StepIOType: Input to the node.
            output: Optional[str]: Output of the node.
            name: Optional[str]: Name of the trace.
            duration_ns: Optional[int]: Duration of the trace in nanoseconds.
            created_at_ns: Optional[int]: Timestamp of the trace's creation.
            metadata: Optional[Dict[str, str]]: Metadata associated with this trace.
            ground_truth: Optional[str]: Ground truth, expected output of the trace.
        Returns:
        -------
            Trace: The created trace.
        """
        trace = Trace(
            input=input,
            output=output or "",
            name=name,
            duration_ns=duration_ns,
            created_at_ns=created_at_ns,
            metadata=metadata,
            ground_truth=ground_truth,
        )
        self.traces.append(trace)
        self.current_parent = trace
        return trace

    def add_single_span_trace(
        self,
        input: LlmStepAllowedIOType,
        output: LlmStepAllowedIOType,
        model: str,
        tools: Optional[List[Dict]] = None,
        name: Optional[str] = None,
        duration_ns: Optional[int] = None,
        created_at_ns: Optional[int] = None,
        metadata: Optional[Dict[str, str]] = None,
        input_tokens: Optional[int] = None,
        output_tokens: Optional[int] = None,
        total_tokens: Optional[int] = None,
        temperature: Optional[float] = None,
        ground_truth: Optional[str] = None,
        status_code: Optional[int] = None,
    ) -> Trace:
        """
        Create a new trace with a single span and add it to the list of traces.

        Parameters:
        ----------
            input: LlmStepAllowedIOType: Input to the node.
            output: LlmStepAllowedIOType: Output of the node.
            model: str: Model used for this span. Feedback from April: Good docs about what model names we use.
            tools: Optional[List[Dict]]: List of available tools passed to LLM on invocation.
            name: Optional[str]: Name of the span.
            duration_ns: Optional[int]: duration_ns of the node in nanoseconds.
            created_at_ns: Optional[int]: Timestamp of the span's creation.
            metadata: Optional[Dict[str, str]]: Metadata associated with this span.
            input_tokens: Optional[int]: Number of input tokens.
            output_tokens: Optional[int]: Number of output tokens.
            total_tokens: Optional[int]: Total number of tokens.
            temperature: Optional[float]: Temperature used for generation.
            ground_truth: Optional[str]: Ground truth, expected output of the workflow.
            status_code: Optional[int]: Status code of the node execution.
        Returns:
        -------
            Trace: The created trace.
        """
        trace = Trace(
            input=input,
            output=output,
        )
        trace.add_llm_span(
            input=input,
            output=output,
            model=model,
            tools=tools,
            name=name,
            duration_ns=duration_ns,
            created_at_ns=created_at_ns,
            metadata=metadata,
            input_tokens=input_tokens,
            output_tokens=output_tokens,
            total_tokens=total_tokens,
            temperature=temperature,
            status_code=status_code,
            ground_truth=ground_truth,
        )
        self.traces.append(trace)
        # Single span traces are automatically concluded so we reset the current parent.
        self.current_parent = None
        return trace

    def add_llm_span(
        self,
        input: LlmStepAllowedIOType,
        output: LlmStepAllowedIOType,
        model: str,
        tools: Optional[List[Dict]] = None,
        name: Optional[str] = None,
        duration_ns: Optional[int] = None,
        created_at_ns: Optional[int] = None,
        metadata: Optional[Dict[str, str]] = None,
        input_tokens: Optional[int] = None,
        output_tokens: Optional[int] = None,
        total_tokens: Optional[int] = None,
        temperature: Optional[float] = None,
        status_code: Optional[int] = None,
    ) -> LlmSpan:
        """
        Add a new llm span to the current parent.

        Parameters:
        ----------
            input: LlmStepAllowedIOType: Input to the node.
            output: LlmStepAllowedIOType: Output of the node.
            model: str: Model used for this span.
            tools: Optional[List[Dict]]: List of available tools passed to LLM on invocation.
            name: Optional[str]: Name of the span.
            duration_ns: Optional[int]: duration_ns of the node in nanoseconds.
            created_at_ns: Optional[int]: Timestamp of the span's creation.
            metadata: Optional[Dict[str, str]]: Metadata associated with this span.
            input_tokens: Optional[int]: Number of input tokens.
            output_tokens: Optional[int]: Number of output tokens.
            total_tokens: Optional[int]: Total number of tokens.
            temperature: Optional[float]: Temperature used for generation.
            status_code: Optional[int]: Status code of the node execution.
        Returns:
        -------
            LlmSpan: The created span.
        """
        if self.current_parent is None:
            raise ValueError("A trace needs to be created in order to add a span.")
        span = self.current_parent.add_llm_span(
            input=input,
            output=output,
            model=model,
            tools=tools,
            name=name,
            duration_ns=duration_ns,
            created_at_ns=created_at_ns,
            metadata=metadata,
            input_tokens=input_tokens,
            output_tokens=output_tokens,
            total_tokens=total_tokens,
            temperature=temperature,
            status_code=status_code,
        )
        return span

    def add_retriever_span(
        self,
        input: StepIOType,
        documents: RetrieverStepAllowedOutputType,
        name: Optional[str] = None,
        duration_ns: Optional[int] = None,
        created_at_ns: Optional[int] = None,
        metadata: Optional[Dict[str, str]] = None,
        status_code: Optional[int] = None,
    ) -> RetrieverSpan:
        """
        Add a new retriever span to the current parent.

        Parameters:
        ----------
            input: StepIOType: Input to the node.
            documents: Union[List[str], List[Dict[str, str]], List[Document]]: Documents retrieved from the retriever.
            name: Optional[str]: Name of the span.
            duration_ns: Optional[int]: duration_ns of the node in nanoseconds.
            created_at_ns: Optional[int]: Timestamp of the span's creation.
            metadata: Optional[Dict[str, str]]: Metadata associated with this span.
            status_code: Optional[int]: Status code of the node execution.
        Returns:
        -------
            RetrieverSpan: The created span.
        """
        if self.current_parent is None:
            raise ValueError("A trace needs to be created in order to add a span.")
        span = self.current_parent.add_retriever_span(
            input=input,
            documents=documents,
            name=name,
            duration_ns=duration_ns,
            created_at_ns=created_at_ns,
            metadata=metadata,
            status_code=status_code,
        )
        return span

    def add_tool_span(
        self,
        input: StepIOType,
        output: StepIOType,
        name: Optional[str] = None,
        duration_ns: Optional[int] = None,
        created_at_ns: Optional[int] = None,
        metadata: Optional[Dict[str, str]] = None,
        status_code: Optional[int] = None,
    ) -> ToolSpan:
        """
        Add a new tool span to the current parent.

        Parameters:
        ----------
            input: StepIOType: Input to the node.
            output: StepIOType: Output of the node.
            name: Optional[str]: Name of the span.
            duration_ns: Optional[int]: duration_ns of the node in nanoseconds.
            created_at_ns: Optional[int]: Timestamp of the span's creation.
            metadata: Optional[Dict[str, str]]: Metadata associated with this span.
            status_code: Optional[int]: Status code of the node execution.
        Returns:
        -------
            ToolSpan: The created span.
        """
        if self.current_parent is None:
            raise ValueError("A trace needs to be created in order to add a span.")
        span = self.current_parent.add_tool_span(
            input=input,
            output=output,
            name=name,
            duration_ns=duration_ns,
            created_at_ns=created_at_ns,
            metadata=metadata,
            status_code=status_code,
        )
        return span

    def add_workflow_span(
        self,
        input: StepIOType,
        output: Optional[StepIOType] = None,
        name: Optional[str] = None,
        duration_ns: Optional[int] = None,
        created_at_ns: Optional[int] = None,
        metadata: Optional[Dict[str, str]] = None,
    ) -> WorkflowSpan:
        """
        Add a workflow span to the current parent. This is useful when you want to create a nested workflow span
        within the trace or current workflow span. The next span you add will be a child of the current parent. To
        move out of the nested workflow, use conclude().

        Parameters:
        ----------
            input: StepIOType: Input to the node.
            output: Optional[StepIOType]: Output of the node. This can also be set on conclude().
            name: Optional[str]: Name of the span.
            duration_ns: Optional[int]: duration_ns of the node in nanoseconds.
            created_at_ns: Optional[int]: Timestamp of the span's creation.
            metadata: Optional[Dict[str, str]]: Metadata associated with this span.
        Returns:
        -------
            WorkflowSpan: The created span.
        """
        if self.current_parent is None:
            raise ValueError("A trace needs to be created in order to add a span.")
        span = self.current_parent.add_workflow_span(
            input=input,
            output=output,
            name=name,
            duration_ns=duration_ns,
            created_at_ns=created_at_ns,
            metadata=metadata,
        )
        self.current_parent = span
        return span

    def conclude(
        self, output: Optional[StepIOType] = None, duration_ns: Optional[int] = None, status_code: Optional[int] = None
    ) -> Optional[StepWithChildSpans]:
        """
        Conclude the current trace or workflow span by setting the output of the current node. In the case of nested
        workflow spans, this will point the workflow back to the parent of the current workflow span.

        Parameters:
        ----------
            output: Optional[StepIOType]: Output of the node.
            duration_ns: Optional[int]: duration_ns of the node in nanoseconds.
            status_code: Optional[int]: Status code of the node execution.
        Returns:
        -------
            Optional[StepWithChildSpans]: The parent of the current workflow. None if no parent exists.
        """
        if self.current_parent is None:
            raise ValueError("No existing workflow to conclude.")
        self.current_parent = self.current_parent.conclude(
            output=output, duration_ns=duration_ns, status_code=status_code
        )
        return self.current_parent



from re import match as re_match
from typing import Union

from pydantic import BaseModel, field_validator


class ScorerNameValidatorMixin(BaseModel):
    name: Union[str, None] = None

    @field_validator("name")
    def validate_name(cls, name: str) -> str:
        """Pydantic field validator for the name field to ensure not empty and no unwanted special characters."""
        if (
            name is not None
        ):  # Allow None to be passed in for optional fields, especially in the case of an UpdateRequest
            if name == "":
                raise ValueError(f"Invalid {cls.__name__}: The name of the scorer cannot be empty.")
            if not re_match(r"^[\w -]+$", name):
                raise ValueError(
                    f"Invalid {cls.__name__}: The scorer cannot contain special characters -- only letters, numbers, space, - and _."
                )
        return name



from typing import Literal, Union

from pydantic import Field
from typing_extensions import Annotated

from galileo_core.schemas.shared.scorers.base_configs import (
    GalileoScorerConfig,
    LunaOrPlusScorerTypeConfig,
    PlusScorerConfig,
    PlusScorerTypeConfig,
    PlusScorerWithNumJudgesConfig,
)
from galileo_core.schemas.shared.scorers.scorer_name import ScorerName


class AgenticSessionSuccessScorer(PlusScorerTypeConfig, PlusScorerWithNumJudgesConfig):
    name: Literal[ScorerName.agentic_session_success] = ScorerName.agentic_session_success


class AgenticWorkflowSuccessScorer(PlusScorerTypeConfig, PlusScorerWithNumJudgesConfig):
    name: Literal[ScorerName.agentic_workflow_success] = ScorerName.agentic_workflow_success


class BleuScorer(GalileoScorerConfig):
    name: Literal[ScorerName.bleu] = ScorerName.bleu


class ChunkAttributionUtilizationScorer(LunaOrPlusScorerTypeConfig, PlusScorerConfig):
    name: Literal[ScorerName.chunk_attribution_utilization] = ScorerName.chunk_attribution_utilization


class CompletenessScorer(LunaOrPlusScorerTypeConfig, PlusScorerWithNumJudgesConfig):
    name: Literal[ScorerName.completeness] = ScorerName.completeness


class ContextAdherenceScorer(LunaOrPlusScorerTypeConfig, PlusScorerWithNumJudgesConfig):
    name: Literal[ScorerName.context_adherence] = ScorerName.context_adherence


class ContextRelevanceScorer(GalileoScorerConfig):
    name: Literal[ScorerName.context_relevance] = ScorerName.context_relevance


class CorrectnessScorer(PlusScorerTypeConfig, PlusScorerWithNumJudgesConfig):
    name: Literal[ScorerName.correctness] = ScorerName.correctness


class GroundTruthAdherenceScorer(PlusScorerTypeConfig, PlusScorerWithNumJudgesConfig):
    name: Literal[ScorerName.ground_truth_adherence] = ScorerName.ground_truth_adherence


class InputPIIScorer(GalileoScorerConfig):
    name: Literal[ScorerName.input_pii] = ScorerName.input_pii


class InputSexistScorer(GalileoScorerConfig):
    name: Literal[ScorerName.input_sexist] = ScorerName.input_sexist


class InputToneScorer(GalileoScorerConfig):
    name: Literal[ScorerName.input_tone] = ScorerName.input_tone


class InputToxicityScorer(GalileoScorerConfig):
    name: Literal[ScorerName.input_toxicity] = ScorerName.input_toxicity


class InstructionAdherenceScorer(PlusScorerTypeConfig, PlusScorerWithNumJudgesConfig):
    name: Literal[ScorerName.instruction_adherence] = ScorerName.instruction_adherence


class OutputPIIScorer(GalileoScorerConfig):
    name: Literal[ScorerName.output_pii] = ScorerName.output_pii


class OutputSexistScorer(GalileoScorerConfig):
    name: Literal[ScorerName.output_sexist] = ScorerName.output_sexist


class OutputToneScorer(GalileoScorerConfig):
    name: Literal[ScorerName.output_tone] = ScorerName.output_tone


class OutputToxicityScorer(GalileoScorerConfig):
    name: Literal[ScorerName.output_toxicity] = ScorerName.output_toxicity


class PromptInjectionScorer(LunaOrPlusScorerTypeConfig, PlusScorerWithNumJudgesConfig):
    name: Literal[ScorerName.prompt_injection] = ScorerName.prompt_injection


class PromptPerplexityScorer(GalileoScorerConfig):
    name: Literal[ScorerName.prompt_perplexity] = ScorerName.prompt_perplexity


class RougeScorer(GalileoScorerConfig):
    name: Literal[ScorerName.rouge] = ScorerName.rouge


class ToolSelectionQualityScorer(PlusScorerTypeConfig, PlusScorerWithNumJudgesConfig):
    name: Literal[ScorerName.tool_selection_quality] = ScorerName.tool_selection_quality


class ToolErrorRateScorer(PlusScorerTypeConfig, PlusScorerConfig):
    name: Literal[ScorerName.tool_error_rate] = ScorerName.tool_error_rate


class UncertaintyScorer(GalileoScorerConfig):
    name: Literal[ScorerName.uncertainty] = ScorerName.uncertainty


GalileoScorersUnion = Union[
    AgenticWorkflowSuccessScorer,
    AgenticSessionSuccessScorer,
    BleuScorer,
    ChunkAttributionUtilizationScorer,
    CompletenessScorer,
    ContextAdherenceScorer,
    ContextRelevanceScorer,
    CorrectnessScorer,
    GroundTruthAdherenceScorer,
    InputPIIScorer,
    InputSexistScorer,
    InputToneScorer,
    InputToxicityScorer,
    InstructionAdherenceScorer,
    OutputPIIScorer,
    OutputSexistScorer,
    OutputToneScorer,
    OutputToxicityScorer,
    PromptInjectionScorer,
    PromptPerplexityScorer,
    RougeScorer,
    ToolErrorRateScorer,
    ToolSelectionQualityScorer,
    UncertaintyScorer,
]


GalileoScorer = Annotated[
    GalileoScorersUnion,
    Field(discriminator="name"),
]



from enum import Enum


class ScorerName(str, Enum):
    agentic_session_success = "agentic_session_success"
    agentic_workflow_success = "agentic_workflow_success"
    bleu = "bleu"
    chunk_attribution_utilization = "chunk_attribution_utilization"
    completeness = "completeness"
    context_adherence = "context_adherence"
    context_relevance = "context_relevance"
    correctness = "correctness"
    ground_truth_adherence = "ground_truth_adherence"
    input_pii = "input_pii"
    input_sexist = "input_sexist"
    input_tone = "input_tone"
    input_toxicity = "input_toxicity"
    instruction_adherence = "instruction_adherence"
    output_pii = "output_pii"
    output_sexist = "output_sexist"
    output_tone = "output_tone"
    output_toxicity = "output_toxicity"
    prompt_injection = "prompt_injection"
    prompt_perplexity = "prompt_perplexity"
    rouge = "rouge"
    tool_error_rate = "tool_error_rate"
    tool_selection_quality = "tool_selection_quality"
    uncertainty = "uncertainty"



from enum import Enum


class NodeType(str, Enum):
    chain = "chain"
    chat = "chat"
    llm = "llm"
    retriever = "retriever"
    tool = "tool"
    agent = "agent"
    workflow = "workflow"
    trace = "trace"



from typing import Any, Dict, List, Literal, Optional, Sequence, Union

from pydantic import Field
from typing_extensions import Annotated

from galileo_core.schemas.shared.workflows.node_type import NodeType
from galileo_core.schemas.shared.workflows.step import (
    BaseStep,
    BaseStepWithChildren,
    LlmStep,
    LlmStepAllowedIOType,
    RetrieverStep,
    RetrieverStepAllowedOutputType,
    StepIOType,
    ToolStep,
)


class SpanWithParentStep(BaseStep):
    parent: Optional["StepWithChildSpans"] = Field(
        default=None, description="Parent node of the current node. For internal use only.", exclude=True
    )


class StepWithChildSpans(BaseStepWithChildren):
    spans: List["Span"] = Field(default_factory=list, description="Child spans.")

    def children(self) -> Sequence[BaseStep]:
        return self.spans

    def add_child(self, *spans: "Span") -> None:
        for span in spans:
            span.parent = self
            self.spans.append(span)

    def add_llm_span(
        self,
        input: LlmStepAllowedIOType,
        output: LlmStepAllowedIOType,
        model: str,
        tools: Optional[Sequence[Dict[str, Any]]] = None,
        name: Optional[str] = None,
        duration_ns: Optional[int] = None,
        created_at_ns: Optional[int] = None,
        metadata: Optional[Dict[str, str]] = None,
        input_tokens: Optional[int] = None,
        output_tokens: Optional[int] = None,
        total_tokens: Optional[int] = None,
        temperature: Optional[float] = None,
        status_code: Optional[int] = None,
        ground_truth: Optional[str] = None,
    ) -> "LlmSpan":
        """
        Add a new llm span to the current workflow.

        Parameters:
        ----------
            input: LlmSpanAllowedIOType: Input to the node.
            output: LlmSpanAllowedIOType: Output of the node.
            model: str: Model used for this span.
            tools: Optional[Sequence[Dict[str, Any]]]: List of available tools passed to LLM on invocation.
            name: Optional[str]: Name of the span.
            duration_ns: Optional[int]: duration_ns of the node in nanoseconds.
            created_at_ns: Optional[int]: Timestamp of the span's creation.
            metadata: Optional[Dict[str, str]]: Metadata associated with this span.
            input_tokens: Optional[int]: Number of input tokens.
            output_tokens: Optional[int]: Number of output tokens.
            total_tokens: Optional[int]: Total number of tokens.
            temperature: Optional[float]: Temperature used for generation.
            status_code: Optional[int]: Status code of the node execution.
        Returns:
        -------
            LlmSpan: The created span.
        """
        span = LlmSpan(
            parent=self,
            input=input,
            output=output,
            model=model,
            tools=tools,
            name=name,
            duration_ns=duration_ns,
            created_at_ns=created_at_ns,
            metadata=metadata,
            input_tokens=input_tokens,
            output_tokens=output_tokens,
            total_tokens=total_tokens,
            temperature=temperature,
            status_code=status_code,
            ground_truth=ground_truth,
        )
        self.add_child(span)
        return span

    def add_retriever_span(
        self,
        input: StepIOType,
        documents: RetrieverStepAllowedOutputType,
        name: Optional[str] = None,
        duration_ns: Optional[int] = None,
        created_at_ns: Optional[int] = None,
        metadata: Optional[Dict[str, str]] = None,
        status_code: Optional[int] = None,
    ) -> "RetrieverSpan":
        """
        Add a new retriever span to the current workflow.

        Parameters:
        ----------
            input: SpanIOType: Input to the node.
            documents: Union[List[str], List[Dict[str, str]], List[Document]]: Documents retrieved from the retriever.
            name: Optional[str]: Name of the span.
            duration_ns: Optional[int]: duration_ns of the node in nanoseconds.
            created_at_ns: Optional[int]: Timestamp of the span's creation.
            metadata: Optional[Dict[str, str]]: Metadata associated with this span.
            status_code: Optional[int]: Status code of the node execution.
        Returns:
        -------
            RetrieverSpan: The created span.
        """
        span = RetrieverSpan(
            parent=self,
            input=input,
            output=documents,
            name=name,
            duration_ns=duration_ns,
            created_at_ns=created_at_ns,
            metadata=metadata,
            status_code=status_code,
        )
        self.add_child(span)
        return span

    def add_tool_span(
        self,
        input: StepIOType,
        output: StepIOType,
        name: Optional[str] = None,
        duration_ns: Optional[int] = None,
        created_at_ns: Optional[int] = None,
        metadata: Optional[Dict[str, str]] = None,
        status_code: Optional[int] = None,
    ) -> "ToolSpan":
        """
        Add a new tool span to the current workflow.

        Parameters:
        ----------
            input: SpanIOType: Input to the node.
            output: SpanIOType: Output of the node.
            name: Optional[str]: Name of the span.
            duration_ns: Optional[int]: duration_ns of the node in nanoseconds.
            created_at_ns: Optional[int]: Timestamp of the span's creation.
            metadata: Optional[Dict[str, str]]: Metadata associated with this span.
            status_code: Optional[int]: Status code of the node execution.
        Returns:
        -------
            ToolSpan: The created span.
        """
        span = ToolSpan(
            parent=self,
            input=input,
            output=output,
            name=name,
            duration_ns=duration_ns,
            created_at_ns=created_at_ns,
            metadata=metadata,
            status_code=status_code,
        )
        self.add_child(span)
        return span

    def add_workflow_span(
        self,
        input: StepIOType,
        output: Optional[StepIOType] = None,
        name: Optional[str] = None,
        duration_ns: Optional[int] = None,
        created_at_ns: Optional[int] = None,
        metadata: Optional[Dict[str, str]] = None,
    ) -> "WorkflowSpan":
        """
        Add a nested workflow span to the workflow. This is useful when you want to create a nested workflow within the
        current workflow. The next span you add will be a child of this workflow. To span out of the nested workflow,
        use conclude_workflow().

        Parameters:
        ----------
            input: SpanIOType: Input to the node.
            output: Optional[SpanIOType]: Output of the node. This can also be set on conclude_workflow().
            name: Optional[str]: Name of the span.
            duration_ns: Optional[int]: duration_ns of the node in nanoseconds.
            created_at_ns: Optional[int]: Timestamp of the span's creation.
            metadata: Optional[Dict[str, str]]: Metadata associated with this span.
        Returns:
        -------
            WorkflowSpan: The created span.
        """
        span = WorkflowSpan(
            parent=self,
            input=input,
            output=output or "",
            name=name,
            duration_ns=duration_ns,
            created_at_ns=created_at_ns,
            metadata=metadata,
        )
        self.add_child(span)
        return span

    def conclude(
        self, output: Optional[StepIOType] = None, duration_ns: Optional[int] = None, status_code: Optional[int] = None
    ) -> Optional["StepWithChildSpans"]:
        """
        Conclude the workflow by setting the output of the current node. In the case of nested workflows, this will
        point the workflow back to the parent of the current workflow.

        Parameters:
        ----------
            output: Optional[SpanIOType]: Output of the node.
            duration_ns: Optional[int]: duration_ns of the node in nanoseconds.
            status_code: Optional[int]: Status code of the node execution.
        Returns:
        -------
            Optional[SpanWithChildren]: The parent of the current workflow. None if no parent exists.
        """
        self.output = output or self.output
        self.status_code = status_code
        if duration_ns is not None:
            self.duration_ns = duration_ns

        if isinstance(self, SpanWithParentStep):
            return self.parent
        return None


class Trace(StepWithChildSpans):
    type: Literal[NodeType.trace] = Field(
        default=NodeType.trace, description="Type of the span. By default, it is set to trace."
    )


class WorkflowSpan(StepWithChildSpans, SpanWithParentStep):
    type: Literal[NodeType.workflow] = Field(
        default=NodeType.workflow, description="Type of the span. By default, it is set to workflow."
    )


class LlmSpan(SpanWithParentStep, LlmStep):
    type: Literal[NodeType.llm] = Field(
        default=NodeType.llm, description="Type of the step. By default, it is set to llm."
    )


class RetrieverSpan(SpanWithParentStep, RetrieverStep):
    type: Literal[NodeType.retriever] = Field(
        default=NodeType.retriever, description="Type of the step. By default, it is set to retriever."
    )


class ToolSpan(SpanWithParentStep, ToolStep):
    type: Literal[NodeType.tool] = Field(
        default=NodeType.tool, description="Type of the step. By default, it is set to tool."
    )


Span = Annotated[Union[WorkflowSpan, LlmSpan, RetrieverSpan, ToolSpan], Field(discriminator="type")]



from datetime import datetime
from io import BufferedReader
from pathlib import Path
from typing import Dict, List, Optional, Union

from pydantic import UUID4, BaseModel, Field, model_validator

from galileo_core.constants.dataset_format import DatasetFormat
from galileo_core.utils.dataset import parse_dataset

DatasetType = Union[Path, str, Dict[str, List[str]], List[Dict[str, str]]]


class BaseDatasetRequest(BaseModel):
    format: DatasetFormat = DatasetFormat.csv
    file_path: Path = Field(exclude=True)

    @model_validator(mode="before")
    def dataset_to_path(cls, values: Dict) -> Dict:
        if "file_path" in values:
            values["file_path"], values["format"] = parse_dataset(values["file_path"])
        return values

    @property
    def files(self) -> Dict[str, BufferedReader]:
        return dict(file=self.file_path.open("rb"))


class UploadDatasetRequest(BaseDatasetRequest):
    file_path: Path = Field(exclude=True)

    @property
    def params(self) -> Dict[str, str]:
        return dict(
            format=self.format.value,
        )


class Dataset(BaseModel):
    id: UUID4
    name: str
    num_rows: Optional[int]
    column_names: Optional[List[str]]
    project_count: int
    created_at: datetime
    updated_at: datetime



from json import dumps, loads
from time import time_ns
from typing import Any, Dict, List, Literal, Optional, Sequence, Union

from pydantic import BaseModel, ConfigDict, Field, ValidationInfo, field_serializer, field_validator
from typing_extensions import Annotated

from galileo_core.schemas.protect.payload import Payload
from galileo_core.schemas.protect.response import Response
from galileo_core.schemas.shared.document import Document
from galileo_core.schemas.shared.message import Message
from galileo_core.schemas.shared.workflows.node_type import NodeType
from galileo_core.utils.json import PydanticJsonEncoder

StepIOType = Union[
    str,
    Document,
    Message,
    Dict[str, Any],
    Sequence[str],
    Sequence[Document],
    Sequence[Message],
    Sequence[Dict[str, str]],
    Sequence[Dict[str, Any]],
]
LlmStepAllowedIOType = Union[str, Dict[str, str], Message, Sequence[str], Sequence[Dict[str, str]], Sequence[Message]]
RetrieverStepAllowedOutputType = Union[Sequence[str], Sequence[Dict[str, str]], Sequence[Document]]


class BaseStep(BaseModel):
    type: NodeType = Field(
        default=NodeType.workflow, description="Type of the step. By default, it is set to workflow."
    )
    input: StepIOType = Field(description="Input to the step.", union_mode="left_to_right")
    output: StepIOType = Field(default="", description="Output of the step.", union_mode="left_to_right")
    name: str = Field(default="", description="Name of the step.", validate_default=True)
    created_at_ns: int = Field(
        default_factory=time_ns, description="Timestamp of the step's creation, as nanoseconds since epoch."
    )
    duration_ns: int = Field(default=0, description="Duration of the step in nanoseconds.")
    metadata: Dict[str, str] = Field(default_factory=dict, description="Metadata associated with this step.")
    status_code: Optional[int] = Field(
        default=None, description="Status code of the step. Used for logging failed/errored steps."
    )
    ground_truth: Optional[str] = Field(default=None, description="Ground truth expected output for the step.")

    model_config = ConfigDict(validate_assignment=True)

    @field_validator("name", mode="before")
    def set_name(cls, value: Optional[str], info: ValidationInfo) -> str:
        return value or info.data["type"]

    @field_validator("created_at_ns", mode="before")
    def set_timestamp(cls, value: Optional[int]) -> int:
        if value is None:
            return time_ns()
        return value

    @field_validator("duration_ns", mode="before")
    def set_duration(cls, value: Optional[int]) -> int:
        if value is None:
            return 0
        return value

    @field_validator("metadata", mode="before")
    def set_metadata(cls, value: Optional[Dict[str, str]]) -> Dict[str, str]:
        if value is None:
            return dict()
        return value

    @field_validator("input", "output", mode="after")
    def validate_input_output_serializable(cls, val: StepIOType) -> StepIOType:
        # Make sure we can dump input/output to json string.
        dumps(val, cls=PydanticJsonEncoder)
        return val

    # We disable the check_fields for this serializer because we want to use the same method for `tools`, which is not
    # a field in the this model but only on the `LlmStep` which is a subclass.
    @field_serializer("input", "output", "tools", when_used="json-unless-none", check_fields=False)
    def serialize_to_str(self, input: StepIOType) -> str:
        return input if isinstance(input, str) else dumps(input, cls=PydanticJsonEncoder)


class BaseStepWithChildren(BaseStep):
    def children(self) -> Sequence[BaseStep]:
        raise NotImplementedError("Please Implement this method")


class StepWithChildren(BaseStepWithChildren):
    steps: List["AWorkflowStep"] = Field(default_factory=list, description="Steps in the workflow.")
    parent: Optional["StepWithChildren"] = Field(
        default=None, description="Parent node of the current node. For internal use only.", exclude=True
    )

    def children(self) -> Sequence[BaseStep]:
        return self.steps

    def add_llm(
        self,
        input: LlmStepAllowedIOType,
        output: LlmStepAllowedIOType,
        model: str,
        tools: Optional[Sequence[Dict[str, Any]]] = None,
        name: Optional[str] = None,
        duration_ns: Optional[int] = None,
        created_at_ns: Optional[int] = None,
        metadata: Optional[Dict[str, str]] = None,
        input_tokens: Optional[int] = None,
        output_tokens: Optional[int] = None,
        total_tokens: Optional[int] = None,
        temperature: Optional[float] = None,
        time_to_first_token_ms: Optional[float] = None,
        status_code: Optional[int] = None,
    ) -> "LlmStep":
        """
        Add a new llm step to the current workflow.

        Parameters:
        ----------
            input: LlmStepAllowedIOType: Input to the node.
            output: LlmStepAllowedIOType: Output of the node.
            model: str: Model used for this step.
            tools: Optional[Sequence[Dict[str, Any]]]: List of available tools passed to LLM on invocation.
            name: Optional[str]: Name of the step.
            duration_ns: Optional[int]: duration_ns of the node in nanoseconds.
            created_at_ns: Optional[int]: Timestamp of the step's creation.
            metadata: Optional[Dict[str, str]]: Metadata associated with this step.
            input_tokens: Optional[int]: Number of input tokens.
            output_tokens: Optional[int]: Number of output tokens.
            total_tokens: Optional[int]: Total number of tokens.
            temperature: Optional[float]: Temperature used for generation.
            time_to_first_token_ms: Optional[float]: Time to first token in milliseconds.
            status_code: Optional[int]: Status code of the node execution.
        Returns:
        -------
            LlmStep: The created step.
        """
        step = LlmStep(
            input=input,
            output=output,
            model=model,
            tools=tools,
            name=name,
            duration_ns=duration_ns,
            created_at_ns=created_at_ns,
            metadata=metadata,
            input_tokens=input_tokens,
            output_tokens=output_tokens,
            total_tokens=total_tokens,
            temperature=temperature,
            time_to_first_token_ms=time_to_first_token_ms,
            status_code=status_code,
        )
        self.steps.append(step)
        return step

    def add_retriever(
        self,
        input: StepIOType,
        documents: RetrieverStepAllowedOutputType,
        name: Optional[str] = None,
        duration_ns: Optional[int] = None,
        created_at_ns: Optional[int] = None,
        metadata: Optional[Dict[str, str]] = None,
        status_code: Optional[int] = None,
    ) -> "RetrieverStep":
        """
        Add a new retriever step to the current workflow.

        Parameters:
        ----------
            input: StepIOType: Input to the node.
            documents: Union[List[str], List[Dict[str, str]], List[Document]]: Documents retrieved from the retriever.
            name: Optional[str]: Name of the step.
            duration_ns: Optional[int]: duration_ns of the node in nanoseconds.
            created_at_ns: Optional[int]: Timestamp of the step's creation.
            metadata: Optional[Dict[str, str]]: Metadata associated with this step.
            status_code: Optional[int]: Status code of the node execution.
        Returns:
        -------
            RetrieverStep: The created step.
        """
        step = RetrieverStep(
            input=input,
            output=documents,
            name=name,
            duration_ns=duration_ns,
            created_at_ns=created_at_ns,
            metadata=metadata,
            status_code=status_code,
        )
        self.steps.append(step)
        return step

    def add_tool(
        self,
        input: StepIOType,
        output: StepIOType,
        name: Optional[str] = None,
        duration_ns: Optional[int] = None,
        created_at_ns: Optional[int] = None,
        metadata: Optional[Dict[str, str]] = None,
        status_code: Optional[int] = None,
    ) -> "ToolStep":
        """
        Add a new tool step to the current workflow.

        Parameters:
        ----------
            input: StepIOType: Input to the node.
            output: StepIOType: Output of the node.
            name: Optional[str]: Name of the step.
            duration_ns: Optional[int]: duration_ns of the node in nanoseconds.
            created_at_ns: Optional[int]: Timestamp of the step's creation.
            metadata: Optional[Dict[str, str]]: Metadata associated with this step.
            status_code: Optional[int]: Status code of the node execution.
        Returns:
        -------
            ToolStep: The created step.
        """
        step = ToolStep(
            input=input,
            output=output,
            name=name,
            duration_ns=duration_ns,
            created_at_ns=created_at_ns,
            metadata=metadata,
            status_code=status_code,
        )
        self.steps.append(step)
        return step

    def add_protect(
        self,
        payload: Payload,
        response: Response,
        created_at_ns: Optional[int] = None,
        metadata: Optional[Dict[str, str]] = None,
        status_code: Optional[int] = None,
    ) -> "ToolStep":
        """
        Add a new protect step to the current workflow.

        Parameters:
        ----------
            payload: Payload: Input to Protect `invoke`.
            response: Response: Output from Protect `invoke`.
            name: Optional[str]: Name of the step.
            created_at_ns: Optional[int]: Timestamp of the step's creation.
            metadata: Optional[Dict[str, str]]: Metadata associated with this step.
            status_code: Optional[int]: Status code of the node execution.
        Returns:
        -------
            ToolStep: The created step.
        """
        step = ToolStep(
            input=payload.model_dump(mode="json"),
            output=response.model_dump(mode="json"),
            name="GalileoProtect",
            duration_ns=response.trace_metadata.response_at - response.trace_metadata.received_at,
            created_at_ns=created_at_ns,
            metadata=metadata,
            status_code=status_code,
        )
        self.steps.append(step)
        return step

    def add_sub_workflow(
        self,
        input: StepIOType,
        output: Optional[StepIOType] = None,
        name: Optional[str] = None,
        duration_ns: Optional[int] = None,
        created_at_ns: Optional[int] = None,
        metadata: Optional[Dict[str, str]] = None,
    ) -> "WorkflowStep":
        """
        Add a nested workflow step to the workflow. This is useful when you want to create a nested workflow within the
        current workflow. The next step you add will be a child of this workflow. To step out of the nested workflow,
        use conclude_workflow().

        Parameters:
        ----------
            input: StepIOType: Input to the node.
            output: Optional[StepIOType]: Output of the node. This can also be set on conclude_workflow().
            name: Optional[str]: Name of the step.
            duration_ns: Optional[int]: duration_ns of the node in nanoseconds.
            created_at_ns: Optional[int]: Timestamp of the step's creation.
            metadata: Optional[Dict[str, str]]: Metadata associated with this step.
        Returns:
        -------
            WorkflowStep: The created step.
        """
        step = WorkflowStep(
            parent=self,
            input=input,
            output=output or "",
            name=name,
            duration_ns=duration_ns,
            created_at_ns=created_at_ns,
            metadata=metadata,
        )
        self.steps.append(step)
        return step

    def add_sub_agent(
        self,
        input: StepIOType,
        output: Optional[StepIOType] = None,
        name: Optional[str] = None,
        duration_ns: Optional[int] = None,
        created_at_ns: Optional[int] = None,
        metadata: Optional[Dict[str, str]] = None,
    ) -> "AgentStep":
        """
        Add a nested agent workflow step to the workflow. This is useful when you want to create a nested workflow
        within the current workflow. The next step you add will be a child of this workflow. To step out of the nested
        workflow, use conclude_workflow().

        Parameters:
        ----------
            input: StepIOType: Input to the node.
            output: Optional[StepIOType]: Output of the node. This can also be set on conclude_workflow().
            name: Optional[str]: Name of the step.
            duration_ns: Optional[int]: duration_ns of the node in nanoseconds.
            created_at_ns: Optional[int]: Timestamp of the step's creation.
            metadata: Optional[Dict[str, str]]: Metadata associated with this step.
        Returns:
        -------
            AgentStep: The created step.
        """
        step = AgentStep(
            parent=self,
            input=input,
            output=output or "",
            name=name,
            duration_ns=duration_ns,
            created_at_ns=created_at_ns,
            metadata=metadata,
        )
        self.steps.append(step)
        return step

    def conclude(
        self, output: Optional[StepIOType] = None, duration_ns: Optional[int] = None, status_code: Optional[int] = None
    ) -> Optional["StepWithChildren"]:
        """
        Conclude the workflow by setting the output of the current node. In the case of nested workflows, this will
        point the workflow back to the parent of the current workflow.

        Parameters:
        ----------
            output: Optional[StepIOType]: Output of the node.
            duration_ns: Optional[int]: duration_ns of the node in nanoseconds.
            status_code: Optional[int]: Status code of the node execution.
        Returns:
        -------
            Optional[StepWithChildren]: The parent of the current workflow. None if no parent exists.
        """
        self.output = output or self.output
        self.status_code = status_code
        if duration_ns is not None:
            self.duration_ns = duration_ns
        return self.parent


class WorkflowStep(StepWithChildren):
    type: Literal[NodeType.workflow] = Field(
        default=NodeType.workflow, description="Type of the step. By default, it is set to workflow."
    )


class ChainStep(StepWithChildren):
    type: Literal[NodeType.chain] = Field(
        default=NodeType.chain, description="Type of the step. By default, it is set to chain."
    )


class LlmStep(BaseStep):
    type: Literal[NodeType.llm] = Field(
        default=NodeType.llm, description="Type of the step. By default, it is set to llm."
    )
    input: LlmStepAllowedIOType = Field(description="Input to the LLM step.", union_mode="left_to_right")
    output: LlmStepAllowedIOType = Field(default="", description="Output of the LLM step.", union_mode="left_to_right")
    tools: Optional[Sequence[Dict[str, Any]]] = Field(
        default=None, description="List of available tools passed to the LLM on invocation."
    )
    model: Optional[str] = Field(default=None, description="Model used for this step.")
    input_tokens: Optional[int] = Field(default=None, description="Number of input tokens.")
    output_tokens: Optional[int] = Field(default=None, description="Number of output tokens.")
    total_tokens: Optional[int] = Field(default=None, description="Total number of tokens.")
    temperature: Optional[float] = Field(default=None, description="Temperature used for generation.")
    time_to_first_token_ms: Optional[float] = Field(default=None, description="Time to first token in milliseconds.")


class RetrieverStep(BaseStep):
    type: Literal[NodeType.retriever] = Field(
        default=NodeType.retriever, description="Type of the step. By default, it is set to retriever."
    )
    input: str = Field(description="Input query to the retriever.")
    output: List[Document] = Field(
        default_factory=list,
        description="Documents retrieved from the retriever. This can be a list of strings or `Document`s.",
    )

    @field_validator("output", mode="before")
    def set_output(cls, value: Union[List[str], List[Dict[str, str]], List[Document]]) -> List[Document]:
        # If the output is stored as a JSON-encoded string, try to parse it.
        if isinstance(value, str):
            try:
                value = loads(value)
            except Exception:
                pass
        if isinstance(value, list):
            if all(isinstance(doc, str) for doc in value):
                parsed = [Document.model_validate(dict(content=doc)) for doc in value]
            elif all(isinstance(doc, dict) for doc in value):
                parsed = [Document.model_validate(doc) for doc in value]
            elif all(isinstance(doc, Document) for doc in value):
                parsed = [Document.model_validate(doc) for doc in value]
            else:
                raise ValueError("Retriever output must be a list of strings, a list of dicts, or a list of Documents.")
            return parsed
        raise ValueError("Retriever output must be a list of strings, a list of dicts or a list of Documents.")


class ToolStep(BaseStep):
    type: Literal[NodeType.tool] = Field(
        default=NodeType.tool, description="Type of the step. By default, it is set to tool."
    )


class AgentStep(StepWithChildren):
    type: Literal[NodeType.agent] = Field(
        default=NodeType.agent, description="Type of the step. By default, it is set to agent."
    )


AWorkflowStep = Annotated[
    Union[WorkflowStep, ChainStep, LlmStep, RetrieverStep, ToolStep, AgentStep], Field(discriminator="type")
]



# mypy: disable-error-code=syntax
# We need to ignore syntax errors until https://github.com/python/mypy/issues/17535 is resolved.
from getpass import getpass
from logging import WARNING, _nameToLevel
from os import environ
from pathlib import Path
from ssl import create_default_context
from time import time
from typing import Any, Optional, Type, TypeVar, Union
from urllib.parse import urljoin
from warnings import warn
from webbrowser import open_new_tab

from jwt import decode as jwt_decode
from pydantic import Field, SecretStr, ValidationInfo, field_serializer, field_validator
from pydantic_core import Url
from pydantic_settings import BaseSettings, SettingsConfigDict

from galileo_core.constants.request_method import RequestMethod
from galileo_core.constants.routes import Routes
from galileo_core.helpers.api_client import ApiClient, SSLContext
from galileo_core.helpers.execution import async_run
from galileo_core.helpers.logger import logger
from galileo_core.schemas.core.user import User

AGalileoConfig = TypeVar("AGalileoConfig", bound="GalileoConfig")


class GalileoConfig(BaseSettings):
    log_level: int = Field(default=WARNING, validate_default=True)
    ssl_context: Union[SSLContext, bool] = Field(
        default=True, validate_default=True, exclude=True, description="SSL context for the API client."
    )

    console_url: Url
    api_url: Optional[Url] = None

    # User auth details.
    username: Optional[str] = None
    password: Optional[SecretStr] = None
    api_key: Optional[SecretStr] = None
    sso_id_token: Optional[SecretStr] = None
    sso_provider: Optional[str] = None

    jwt_token: Optional[SecretStr] = None
    current_user: Optional[str] = None

    # Validated API client. This is set as a part of the initialization.
    # We set the API URL and JWT token on the API client and make a request to
    # confirm that the user is logged in successfully.
    # This is set as an exclude field to avoid serializing it to the config file.
    validated_api_client: Optional[ApiClient] = Field(
        default=None,
        validate_default=True,
        exclude=True,
        description="Validated API client that is with the user's JWT token after a successful login.",
    )

    # Config file for this project.
    config_filename: str = "galileo-config.json"

    model_config = SettingsConfigDict(
        # Allow loading from environment variables.
        env_prefix="GALILEO_",
        # Allow unknown fields when loading from a config file.
        extra="allow",
    )

    @field_validator("ssl_context", mode="before")
    def set_ssl_context(cls, value: Union[SSLContext, bool]) -> Union[SSLContext, bool]:
        # Set to default value.
        ssl_cert_file = environ.get("SSL_CERT_FILE")
        ssl_cert_dir = environ.get("SSL_CERT_DIR")
        if value is False:
            warn("SSL is disabled. This is not recommended for production use.", category=UserWarning)
            logger.debug("SSL validation is disabled.")
        elif ssl_cert_file or ssl_cert_dir:
            logger.debug(f"SSL context is set to custom cert file ({ssl_cert_file}) or directory ({ssl_cert_dir}).")
            value = create_default_context(cafile=ssl_cert_file, capath=ssl_cert_dir)
        else:
            logger.debug("SSL context is set to the default context.")
            value = True
        return value

    @property
    def config_file(self) -> Path:
        return Path().home().joinpath(".galileo", self.config_filename)

    @field_validator("log_level", mode="before")
    def set_log_level(cls, value: Optional[Union[int, str]]) -> int:
        """
        Set the log level for the logger.

        We allow setting the log level using the string representation of the log level
        or the integer representation of the log level.

        By default, we set the log level to `WARNING`.

        Parameters
        ----------
        value : Optional[Union[int, str]]
            Log level to set.

        Returns
        -------
        int
            Log level set.
        """
        if isinstance(value, str):
            value = value.upper()
            log_level = _nameToLevel.get(value, WARNING)
        elif isinstance(value, int):
            log_level = value
        else:
            log_level = WARNING
        logger.setLevel(log_level)
        return log_level

    @field_validator("console_url", mode="before")
    def ensure_https_console_url(cls, value: str) -> str:
        """
        Ensure that the console URL is an HTTPS URL.

        Parameters
        ----------
        value : str
            Console URL to validate.

        Returns
        -------
        str
            Validated console URL.
        """
        if value and not (value.startswith("https") or value.startswith("http")):
            value = f"https://{value}"
        return value

    @field_validator("api_url", mode="before")
    def set_api_url(cls, api_url: Optional[Union[str, Url]], info: ValidationInfo) -> Url:
        """
        Set the API URL if it's not already set.

        We can set the API URL in the following ways:
        1. If the API URL is provided, use it.
        2. If the console URL matches `localhost` or `127.0.0.1`, use `http://localhost:8088`.
        3. If the API URL is not provided, use the console URL to generate the API URL
        by replacing `console` with `api`.

        Once we determine the API URL, we make a request to the healthcheck endpoint to
        validate that it is reachable.

        Parameters
        ----------
        api_url : Optional[Union[str, Url]]
            API URL to use. If not provided, we generate it from the console URL.
        info : ValidationInfo
            Pydantic validation info object.

        Returns
        -------
        Url
            API URL to use.

        Raises
        ------
        ValueError
            If the console URL is not set.
        ValueError
            If the API URL can't be generated.
        """
        if api_url is None:
            console_url = info.data.get("console_url")
            if console_url is None:
                raise ValueError(
                    "Console URL is required. Please set the environment variable "
                    "`GALILEO_CONSOLE_URL` to your Galileo console URL."
                )
            else:
                console_url = console_url.unicode_string() if isinstance(console_url, Url) else console_url
            # Local dev.
            if any(["localhost" in console_url, "127.0.0.1" in console_url]):
                api_url = "http://localhost:8088"
            elif "app.galileo.ai" in console_url:
                api_url = "https://api.galileo.ai"
            else:
                api_url = console_url.replace("console", "api")
        if api_url is None:
            raise ValueError("API URL is required.")
        else:
            async_run(
                ApiClient.make_request(
                    RequestMethod.GET,
                    base_url=str(api_url),
                    endpoint=Routes.healthcheck,
                    ssl_context=info.data.get("ssl_context", True),
                )
            )
        return Url(api_url) if isinstance(api_url, str) else api_url

    @staticmethod
    def get_token_from_ui(console_url: Url) -> str:
        """
        Given a console URL, open the token generation page in the browser and prompt the user to enter the token
        generated.

        Parameters
        ----------
        console_url : HttpUrl
            Console URL to use.

        Returns
        -------
        str
            JWT token generated by the user.
        """
        token_url = urljoin(console_url.unicode_string(), Routes.get_token)
        logger.info(f"🔐 Opening {token_url} to generate a new token.")
        try:
            open_new_tab(token_url)
        except Exception:
            # If we can't open the browser, that's fine.
            pass
        finally:
            print(f"Go to {token_url} to generate a new token.")
            return getpass("🔐 Enter your token:")

    @staticmethod
    def get_jwt_token(
        console_url: Optional[Url] = None,
        api_url: Optional[Url] = None,
        api_key: Optional[SecretStr] = None,
        username: Optional[str] = None,
        password: Optional[SecretStr] = None,
        sso_id_token: Optional[SecretStr] = None,
        sso_provider: Optional[str] = None,
        ssl_context: Union[SSLContext, bool] = True,
    ) -> SecretStr:
        """
        Get the JWT token for the user.

        1. If an API key is provided, log in with the API key.
        2. If a username and password are provided, log in with the username and password.
        3. If no credentials are provided, attempt to log in with a token from the UI.

        Parameters
        ----------
        console_url : Optional[HttpUrl], optional
            Console URL, by default None
        api_url : Optional[HttpUrl], optional
            API URL, by default None
        api_key : Optional[SecretStr], optional
            API key, by default None
        username : Optional[str], optional
            Username, by default None
        password : Optional[SecretStr], optional
            Password, by default None
        sso_id_token : Optional[SecretStr], optional
            SSO ID token, by default None
        sso_provider : Optional[str], optional
            SSO provider, by default None
        ssl_context : Union[SSLContext, bool], optional
            SSL context, by default True

        Returns
        -------
        SecretStr
            JWT token for the user, if successful, as a secret string.

        Raises
        ------
        AssertionError
            If the console URL is not provided.
        AssertionError
            If the API URL is not provided.
        """

        token_data = dict()
        assert console_url is not None, "Console URL is required."
        assert api_url is not None, "API URL is required."
        if api_key:
            logger.debug("Logging in with API key.")
            token_data = async_run(
                ApiClient.make_request(
                    RequestMethod.POST,
                    base_url=api_url.unicode_string(),
                    endpoint=Routes.api_key_login,
                    json=dict(api_key=api_key.get_secret_value()),
                    ssl_context=ssl_context,
                )
            )
            logger.debug("Logged in with API key.")
        elif username and password:
            logger.debug("Logging in with username and password.")
            token_data = async_run(
                ApiClient.make_request(
                    RequestMethod.POST,
                    base_url=api_url.unicode_string(),
                    endpoint=Routes.username_login,
                    data=dict(
                        username=username,
                        password=password.get_secret_value(),
                        auth_method="email",
                    ),
                    ssl_context=ssl_context,
                )
            )
            logger.debug("Logged in with username and password.")
        elif sso_id_token and sso_provider:
            logger.debug("Logging in with SSO ID token.")
            token_data = async_run(
                ApiClient.make_request(
                    RequestMethod.POST,
                    base_url=api_url.unicode_string(),
                    endpoint=Routes.social_login,
                    json=dict(id_token=sso_id_token.get_secret_value(), provider=sso_provider),
                    ssl_context=ssl_context,
                )
            )
            logger.debug("Logged in with SSO ID token.")
        if (jwt_token := token_data.get("access_token")) is None:
            logger.debug("No credentials found. Attempting to log in with UI token.")

            jwt_token = GalileoConfig.get_token_from_ui(console_url)
            logger.debug("Logged in with access token from UI.")
        logger.debug("JWT token received and set.")
        return SecretStr(jwt_token)

    @field_validator("jwt_token", mode="before")
    def set_jwt_token(cls, value: Optional[Union[str, SecretStr]], info: ValidationInfo) -> SecretStr:
        """
        Set the JWT token for the user.

        Parameters
        ----------
        value : Optional[str]
            JWT token to set for the user.
        info : ValidationInfo
            Pydantic validation info object.

        Returns
        -------
        SecretStr
            JWT token as a secret string.
        """
        if value is None:
            console_url, api_url, api_key, username, password, sso_id_token, sso_provider, ssl_context = (
                info.data.get("console_url"),
                info.data.get("api_url"),
                info.data.get("api_key"),
                info.data.get("username"),
                info.data.get("password"),
                info.data.get("sso_id_token"),
                info.data.get("sso_provider"),
                info.data.get("ssl_context", True),
            )
            value = cls.get_jwt_token(
                console_url, api_url, api_key, username, password, sso_id_token, sso_provider, ssl_context
            )
        assert value is not None, "JWT token is required."
        return SecretStr(value) if isinstance(value, str) else value

    @field_validator("validated_api_client", mode="before")
    def set_validated_api_client(cls, validated_api_client: Optional[ApiClient], info: ValidationInfo) -> ApiClient:
        """
        Set the validated API client.

        This method sets an API client with the validated API URL and JWT token. As a
        part of the validation process, we make a request to get the current user's email
        address to confirm that the user is logged in successfully.

        Parameters
        ----------
        validated_api_client : Optional[ApiClient]
            API client to set.
        info : ValidationInfo
            Pydantic validation info object.

        Returns
        -------
        ApiClient
            Validated API client.
        """
        console_url, api_url, jwt_token, ssl_context = (
            info.data.get("console_url"),
            info.data.get("api_url"),
            info.data.get("jwt_token"),
            info.data.get("ssl_context", False),
        )
        assert api_url is not None, "API URL is required."
        assert jwt_token is not None, "JWT token is required."
        validated_api_client = ApiClient(host=api_url, jwt_token=jwt_token, ssl_context=ssl_context)
        # Get the current user to confirm that the user is logged in successfully.
        current_user_dict = validated_api_client.request(RequestMethod.GET, path=Routes.current_user)
        current_user = User.model_validate(current_user_dict)
        logger.debug("Logged in successfully.")
        print(f"👋 You have logged into 🔭 Galileo ({console_url}) as {current_user.email}.")
        return validated_api_client

    @field_serializer("password", "jwt_token", "api_key", "sso_id_token", when_used="json-unless-none")
    def serialize_secrets(self, value: SecretStr) -> str:
        """Serialize secret strings to their secret values."""
        return value.get_secret_value()

    def write(self) -> None:
        """
        Write the config object to a file.

        This is only used as a backup for debugging and never read from to set current values.
        """
        self.config_file.parent.mkdir(parents=True, exist_ok=True)
        self.config_file.write_text(self.model_dump_json(exclude_none=True))

    def refresh_jwt_token(self) -> None:
        """Refresh token if not present or expired."""
        # Check to see if our token is expired or (will expire in the next 5 minutes) before making a request and
        # refresh token if it's expired.
        if self.jwt_token:
            claims = jwt_decode(self.jwt_token.get_secret_value(), options={"verify_signature": False})
            if claims.get("exp", 0) <= (time() + 300):
                logger.debug("JWT token is invalid, refreshing.")
                self.jwt_token = self.get_jwt_token(
                    self.console_url,
                    self.api_url,
                    self.api_key,
                    self.username,
                    self.password,
                    self.sso_id_token,
                    self.sso_provider,
                )
            else:
                logger.debug("JWT token is still valid, not refreshing.")
        # If no token is present, log in.
        else:
            logger.debug("JWT token not found, getting a new one.")
            self.jwt_token = self.get_jwt_token(
                self.console_url,
                self.api_url,
                self.api_key,
                self.username,
                self.password,
                self.sso_id_token,
                self.sso_provider,
            )

    @property
    def api_client(self) -> ApiClient:
        """
        Get the API client.

        We refresh the JWT token if it's expired and set the JWT token on the API client
        if it's different from the one set on the API client during the config
        initialization.

        Returns
        -------
        ApiClient
            Validated API client.
        """
        assert self.validated_api_client is not None, "API client must be set before accessing it."
        self.refresh_jwt_token()
        if self.jwt_token and self.validated_api_client.jwt_token != self.jwt_token:
            self.validated_api_client.jwt_token = self.jwt_token
        return self.validated_api_client

    @classmethod
    def _get(cls: Type[AGalileoConfig], existing_config: Optional[AGalileoConfig], **kwargs: Any) -> "AGalileoConfig":
        """
        Internal method to get the config object.

        This method is used to set the config object from the global variable (passed in as an arg) or from the kwargs
        or environment variables. We use this so that we can maintain independent global variables for different
        sub-classes of the BaseConfig class in their respective modules, while still keeping the interaction predictable
        and consistent.
        """
        if kwargs:
            logger.debug("Setting config from arguments and environment variables.")
            return cls(**kwargs)
        elif existing_config is None:
            logger.debug("Setting config from environment variables.")
            return cls()
        else:
            # Ignore the type here because we know that _config is not None and is an object of the
            # GalileoConfig class or its sub-classes.
            logger.debug("Config is already set.")
            return existing_config  # type: ignore[return-value]

    @classmethod
    def get(cls: Type[AGalileoConfig], **kwargs: Any) -> "AGalileoConfig":
        """
        Get the config object from the global variable or set it if it's not already set from the kwargs or environment
        variables.

        This method should be implemented in the sub-classes to set the config object for the sub-class.
        """
        global _config
        # Ignore the type here because we know that _config is an object of the BaseConfig class or its sub-classes.
        _config = cls._get(_config, **kwargs)  # type: ignore[arg-type]
        return _config

    def reset(self) -> None:
        """
        Reset the credentials stored in the config object.

        Sub-classes can extend this method to reset additional fields.
        """
        self.username = None
        self.password = None
        self.api_key = None
        self.sso_id_token = None
        self.sso_provider = None

        self.jwt_token = None
        self.current_user = None
        self.validated_api_client = None

        global _config
        _config = None

        self.write()

    def logout(self) -> None:
        """Logout the user by resetting the credentials and printing a message."""
        self.reset()
        print(f"👋 You have logged out of 🔭 Galileo ({self.console_url}).")


# Global config object that is used to store the config object after the first load.
_config: Optional[GalileoConfig] = None
